"""
Python interface to the RELIC cryptographic library.
"""
import atexit, ctypes, sys
from common import *
from os import path

# NOTE: This module was designed, built, and tested assuming that the RELIC
#       C library was built on a 64-bit OSX platform. This module further
#       assumes the follow RELIC configuration settings
#
# ALLOC = AUTO
# WORD = 64
# BN_MAGNI = DOUBLE
# BN_SIZE = 18

# Known file extensions for dynamically-linked libraries (shared objects)
# on various OS's that we've tested and support.

# OSX
if sys.platform == "darwin":
    ext = "dylib"

# Linux
elif sys.platform.startswith("linux"):
    ext = "so"

# Windows - this has never been tested
elif sys.platform == "win32":
    ext = "dll"

else:
    raise Exception("Unsupported operating system: only OSX, Linux, and "\
        "Windows are currently supported.")


# Build the path to the librelic DLL
name = "librelic." + ext
moduleDirectory = path.dirname(__file__)
libPath = path.join(moduleDirectory,  "lib", name)

# Load the relic library
librelic = ctypes.cdll.LoadLibrary(libPath)

# Initialize the RELIC core (memory allocation, error handling, and  
# other internal state)
if librelic.core_init() != 0:
    raise Exception("Could not initialize RELIC core")

# Set the pairing based curve (PC) parameters.
if librelic.pc_param_set_any_abi() != 0:
    raise Exception("Could not set PBC parameters")


@atexit.register
def cleanup():
    """
    Relic library clean-up routine. Registered to be called on module exit.
    """
    librelic.core_clean()




#!/usr/bin/eval python

from testcommon import *
from pbc import *
from timeit import timeit
from unittest import TestCase, SkipTest
import unittest


class PbcTests(TestCase):
    """
    Tests for the pairing-based crypto classes and related module functions.
    """
    def testPair(self):
        """
        Test pairing funciton by checking the multiplicative homomorphic 
        property.
        """
        r = randomZ()
        p = randomG1()
        q = randomG2()

        t1 = pair(p*r,q)
        t2 = pair(p,q*r)

        self.assertEqual(t1, t2)


    def testRandomG1(self):
        """
        Grabs random elements from G1 an ensure there are no duplicates. 
        """
        randomNoRepeat(randomG1)


    def testRandomG2(self):
        """
        Grabs random elements from G2 an ensure there are no duplicates. 
        """
        randomNoRepeat(randomG2)


    def testRandomZ(self):
        """
        Grab @n random elements from Z an ensure there are no duplicates. 
        """
        randomNoRepeat(randomZ)


class HashTestBase(TestCase):
    """
    Base class for testing hashG1 and hashG2 functions.
    """
    def setUp(self):
        raise unittest.SkipTest("Base class")


    def testHashType(self):
        """
        Tests the type of the hash output
        """
        m = "This is a very long string. It makes a nice test message."
        h = self.hashfunc(m)
        self.assertTrue( isinstance(h, self.expectedType) )


    def testHashSame(self):
        """
        Each hash of the same string must give the same result
        """
        # We want to messages that give the same bytes, but are not identical
        # objects.
        m1 = "This is a very long string. It makes a nice test message."
        m2 = "This is a very long string." + " It makes a nice" + " test message."

        # Ensure we have two unique objects with identical values
        self.assertEqual(m1, m2)
        self.assertIsNot(m1, m2)

        h1 = self.hashfunc(m1)
        h2 = self.hashfunc(m2)
        self.assertEqual(h1, h2)


    def testHashDifferent(self):
        """
        Hashes of different strings must give unique results.
        """
        # We want to messages that give the same bytes, but are not identical
        # objects.
        m1 = "This is a test hash input"
        m2 = "This is a distinctly different hash inputs from m1."

        h1 = self.hashfunc(m1)
        h2 = self.hashfunc(m2)
 
        # Ensure the results are distinct.
        self.assertNotEqual(h1, h2)


    def testHashMultiple(self):
        """
        Test hashing multiple arguments.
        """
        m1 = "First hash input"
        m2 = 981723509871230498709871234
        m3 = "Third input"

        h1a = self.hashfunc(m1, m2, m3)
        h1b = self.hashfunc(m1, m2, m3)
        h2 = self.hashfunc(m2, m1, m3)
        h3 = self.hashfunc(m1, m2)

        # Same inputs should give the same result
        self.assertEqual(h1a, h1b)

        # Different inputs (or order) should give different results.
        self.assertNotEqual(h1a, h2)
        self.assertNotEqual(h2, h3)
        self.assertNotEqual(h1a, h3)


class PbcHashG1Test(HashTestBase):
    def setUp(self):
        self.hashfunc = hashG1
        self.expectedType = G1Element


class PbcHashG2Test(HashTestBase):
    def setUp(self):
        self.hashfunc = hashG2
        self.expectedType = G2Element



import base64

class PbcSerialBase(TestCase):
    """
    Base class for testing the de/serialization of G1, G2, and Gt elements.
    """
    def setUp(self):
        raise unittest.SkipTest("Base class")


    def assertNotEmpty(self, x):
        """
        Ensures that @x is not None and len(x) > 0
        """
        self.assertTrue(x is not None and len(x) > 0)


    def fullCycle(self, x, compress=False):
        """
        Tests de/serialization by running full cycle and comparing the result 
        to the original element.
        """
        b = self.serialize(x, compress)
        x2 = self.deserialize(b, compress)
        self.assertEqual(x, x2)


    def testSerializeNonEmpty(self, compress=False):
        """
        Tests that encoding an element delivers a non-empty list of bytes.
        """
        self.assertNotEmpty( self.serialize(self.randomElement(), compress) )


    def testSerializeNonEmptyCompressed(self):
        """
        Tests that encoding an element with compression delivers a non-empty 
        list of bytes.
        """
        self.testSerializeNonEmpty(compress=True)


    def testCompressed(self):
        """
        Tests that serializing with compression results in a shorter result than
        the uncompressed routine.
        """
        x = self.randomElement()
        b1 = self.serialize(x, False)
        b2 = self.serialize(x, True)

        self.assertNotEmpty(b1)
        self.assertNotEmpty(b2)
        self.assertTrue(len(b2) < len(b1))


    def testFullCycle(self, n=10, compress=False):
        func = lambda: self.fullCycle(self.randomElement(), compress)
        repeat(func, n)


    def testFullCycleCompressed(self):
        self.testFullCycle(compress=True)


    def testFullCycleEncoded(self, compress=False):
        """
        Tests full cycle serialization + b64 encoding.
        """
        # serialize and encode a random element
        x = self.randomElement()
        b = bytearray( self.serialize(x, compress) )
        encoded = base64.urlsafe_b64encode(b)

        # Decode the result and verify that the underlying bytes still match
        decodedBytes = base64.urlsafe_b64decode(encoded)
        self.assertEqual(b, decodedBytes)

        # Deserialize and compare against the original element
        x2 = self.deserialize(decodedBytes)
        self.assertEqual(x, x2)


    def testFullCycleEncodedCompressed(self, compress=False):
        """
        Tests full cycle serialization with compression and encoding.
        """
        self.testFullCycleEncoded(compress=True)


class G1SerializeTests(PbcSerialBase):
    def setUp(self):
        self.randomElement = randomG1
        self.serialize = serializeG1
        self.deserialize = deserializeG1


class G2SerializeTests(PbcSerialBase):
    def setUp(self):
        self.randomElement = randomG2
        self.serialize = serializeG2
        self.deserialize = deserializeG2


class GtSerializeTests(PbcSerialBase):
    def setUp(self):
        self.randomElement = randomGt
        self.serialize = serializeGt
        self.deserialize = deserializeGt


# Run!
if __name__ == '__main__':
    unittest.main()


"""
Pythia's Verifiable, Partially-Oblivious Pseudorandom Function (POP) protocol 
constructed using the BN-256 pairing-based curves provided by the RELIC library.
Also includes serialization and encoding routines for elements that are commonly
transmitted.
"""
from pbc import *
from prf import *

def eval(w,t,x,msk,s):
    """
    Pythia server-side computation of intermediate PRF output.
    @w: ensemble key selector (e.g. webserver ID)
    @t: tweak (e.g. user ID)
    @x: blinded message (element of G1)
    @msk: Pythia server's master secret key
    @s: state value from Pythia server's key table
    @returns: (y, kw, tTile)
     where: y: intermediate result
            kw: secret key bound to w (needed for proof)
            tTilde: hashed tweak (needed for proof)
    """
    # Construct the key
    kw = genKw(w,msk,s)

    # Multiply x by kw (it's fastest this way), hash the tweak, and compute
    # the pairing.
    tTilde = hashG2(t)
    y = pair(x*kw, tTilde)
    return y,kw,tTilde


def prove(x,tTilde,kw,y):
    """
    Generate a zero-knowledge proof that DL(Q*kw) == DL(e(x,tTilde)^kw) where
    <Q> = G1.
    @x: Blinded message from client request.
    @tTilde: HG2(t), element of G2
    @kw: secret key derived from w
    @y: intermediate result from eval function. element of Gt
    """
    # Verify types
    assertType(x, G1Element)
    assertType(tTilde, G2Element)
    assertType(y, GtElement)

    # Compute the proof.
    beta = pair(x,tTilde)
    Q = generatorG1()
    p = Q*kw
    v = randomZ(orderGt())
    t1 = Q*v
    t2 = beta**v

    t1.normalize()

    c = hashZ(Q,p,beta,y,t1,t2)
    u = (v-(c*kw)) % orderGt()
    return (p,c,u)


def verify(x, t, y, pi, errorOnFail=True):
    """
    Verifies a zero-knowledge proof where p \in G1.
    @errorOnFail: Raise an exception if the proof does not hold.
    """
    # Unpack the proof
    p,c,u = pi

    # Verify types
    assertType(x, G1Element)
    assertType(y, GtElement)
    assertType(p, G1Element)
    assertScalarType(c)
    assertScalarType(u)

    # TODO: beta can be pre-computed while waiting for a server response.
    Q = generatorG1()
    beta = pair(x,hashG2(t))

    # Recompute c'
    t1 = Q*u + p*c 
    t2 = beta**u * y**c

    t1.normalize()

    cPrime = hashZ(Q,p,beta,y,t1,t2)

    # Check computed @c' against server's value @c
    if cPrime == c:
        return True

    if errorOnFail:
        raise Exception("zero-knowledge proof failed verification.")
    else:
        return False


def blind(m, hashfunc=hashG1):
    """
    Blinds an arbitrary string or byte array @m using an ephemeral key @r
    that can be used to deblind. Computes: x = H(x)^r
    @returns (1/r,x)
    """
    # Find r with a suitable inverse in Gt
    rInv = None
    while not rInv:
        r = randomZ()
        rInv = inverse(r, orderGt())

    return rInv, hashfunc(m) * r


def deblind(rInv,y):
    """
    Removes blinding using ephemeral key @rInv on (intermediate result) 
    @y \in Gt.
    """
    # Verify types, then deblind using the values provided.
    assertScalarType(rInv)
    assertType(y, GtElement)
    return y ** rInv


# Decode/deserialize elements by name
unwrapX = unwrapG1
unwrapY = unwrapGt
unwrapP = unwrapG1
unwrapC = unwrapLong
unwrapU = unwrapLong



"""
Common testing routines
"""

import random, string

def randomstr(n=64):
    """
    Generate a random string of length @n.
    """
    pop = string.digits + string.lowercase + string.uppercase
    return ''.join(random.choice(pop) for _ in range(n))


def randomNoRepeat(func, n=100):
    """
    Grab @n elements using @func and ensure there are no duplicates 
    according to their string representations.
    """
    seen = set()
    for _ in range(n):
        g = str(func())

        # Ensure this element has been seen previously
        if g in seen:
            raise Exception("Unexpected duplicate: {}".format(g))

        seen.add(g)


def repeat(func, n=1000):
    """
    Call @func @n times.
    """
    for _ in range(n):
        func()



"""
Pythia's Verifiable, Partially-Oblivious Pseudorandom Function (POP) protocol 
constructed using the BN-256 pairing-based curves provided by the RELIC library.
Also includes serialization and encoding routines for elements that are commonly
transmitted.
"""
from pbc import *

def eval(w,t,x,msk,s):
    """
    Pythia server-side computation of intermediate PRF output.
    @w: ensemble key selector (e.g. webserver ID)
    @t: tweak (e.g. user ID)
    @x: blinded message (element of G1)
    @msk: Pythia server's master secret key
    @s: state value from Pythia server's key table
    @returns: (y, kw, tTile)
     where: y: intermediate result
            kw: secret key bound to w (needed for proof)
            tTilde: hashed tweak (needed for proof)
    """
    kw = genKw(w,msk,s)

    # TODO: Return cached values for precomputation
    tTilde = hashG2(t)
    y = pair(x*kw, tTilde)
    return y,kw,tTilde


def genKw(w,msk,z):
    """
    Generates key Kw using key-selector @w, master secret key @msk, and
    table value @z.
    @returns Kw as a BigInt.
    """
    # Hash inputs into a string of bytes
    TAG_KW = "TAG_PYTHIA_KW"
    b = hmac(TAG_KW, msk, z + w)

    # Convert the string into a long value (no larger than the order of Gt),
    # then return a BigInt value.
    return BigInt(longFromString(b) % long(orderGt()))


def prove(x,tTilde,kw,y):
    return proveGt(x,tTilde,kw,y)

def verify(x, tTilde, y, pi, errorOnFail=True):
    return verifyGt(x, tTilde, y, pi, errorOnFail)

@profile
def proveGt(x,tTilde,kw,y):
    """
    Generate a zero-knowledge proof that DL(g^kw) == DL(e(x,t)^kw) where
    g,e(..) \in Gt.
    @return pi = (p,c,u)
    """
    # Verify types
    assertType(x, G1Element)
    assertType(tTilde, G2Element)

    # Compute the proof.
    beta = pair(x,tTilde)
    g = generatorGt()
    p = g**kw
    v = randomZ(orderGt())
    t1 = g**v
    t2 = beta**v

    c = hashZ(g,p,beta,y,t1,t2)
    u = (v- (c*kw)) % orderGt()
    return (p,c,u)

@profile
def proveG1(x,tTilde,kw,y):
    """
    Generate a zero-knowledge proof that DL(Q*kw) == DL(e(x,tTilde)^kw) where
    <Q> = G1.
    """
    # Verify types
    assertType(x, G1Element)
    assertType(tTilde, G2Element)

    # Compute the proof.
    beta = pair(x,tTilde)
    Q = generatorG1()
    p = Q*kw
    v = randomZ(orderGt())
    t1 = Q*v
    t2 = beta**v

    t1.normalize()

    c = hashZ(Q,p,beta,y,t1,t2)
    u = (v-(c*kw)) % orderGt()
    return (p,c,u)

@profile
def verifyG1(x, tTilde, y, pi, errorOnFail=True):
    """
    Verifies a zero-knowledge proof where p \in G1.
    @errorOnFail: Raise an exception if the proof does not hold.
    """
    # Unpack the proof
    p,c,u = pi

    # Verify types
    assertType(x, G1Element)
    assertType(tTilde, G2Element)
    assertType(y, GtElement)
    assertType(p, G1Element)

    # TODO: beta can be pre-computed while waiting for a server response.
    Q = generatorG1()
    beta = pair(x,tTilde)

    # Recompute c'
    t1 = Q*u + p*c 
    t2 = beta**u * y**c

    t1.normalize()

    cPrime = hashZ(Q,p,beta,y,t1,t2)

    # Check computed @c' against server's value @c
    if cPrime == c:
        return True

    if errorOnFail:
        raise Exception("zero-knowledge proof failed verification.")
    else:
        return False

@profile
def verifyGt(x, tTilde, y, pi, errorOnFail=True):
    """
    Verifies a zero-knowledge proof. 
    @x: Blinded message, G1Element, x = HG1(m)*r
    @tTilde: hashed tweak, G2Element, t~ = HG2(t)
    @y: server response (intermediate result), GtElement
    @pi:  pi = (p, c, u), zero-knowledge proof from server, 
      p = g^kw, GtElement; c,u integer values
    @errorOnFail: Rasise an exception if the proof does not hold.
    """
    # Unpack the proof
    p,c,u = pi

    # Verify types
    assertType(x, G1Element)
    assertType(tTilde, G2Element)
    assertType(y, GtElement)
    assertType(p, GtElement)

    # TODO: This can be pre-computed while waiting for a server response.
    g = generatorGt()
    beta = pair(x,tTilde)

    # Recompute c'
    t1 = g**u * p**c 
    t2 = beta**u * y**c

    cPrime = hashZ(g,p,beta,y,t1,t2)

    # Check computed @c' against server's value @c
    if cPrime == c:
        return True

    if errorOnFail:
        raise Exception("zero-knowledge proof failed verification.")
    else:
        return False


def blind(m, hashfunc=hashG1):
    """
    Blinds an arbitrary string or byte array @m using an ephemeral key @r
    that can be used to deblind. Computes: x = H(x)^r
    @returns (1/r,x)
    """
    # Find r with a suitable inverse in Gt
    rInv = None
    while not rInv:
        r = randomZ()
        rInv = inverse(r, orderGt())

    return rInv, hashfunc(m) * r


def deblind(rInv,y):
    """
    Removes blinding using ephemeral key @r on (intermediate result) @y \in Gt.
    """
    # Find the multiplicative inverse of @r in Gt.
    return y ** rInv


import base64 
def _wrap(x, serializeFunc, encodeFunc=base64.urlsafe_b64encode, compress=True):
    """
    Wraps an element @x by serializing and then encoding the resulting bytes.
    """
    return encodeFunc(serializeFunc(x, compress))

def _unwrap(x, deserializeFunc, decodeFunc=base64.urlsafe_b64decode, compress=True):
    """
    Unwraps an element @x by decoding and then deserializing
    """
    return deserializeFunc(decodeFunc(x), compress)

def _wrapG1(x):
    return _wrap(x, serializeG1)

def _unwrapG1(x):
    return _unwrap(x, deserializeG1)

def _wrapGt(x):
    return _wrap(x, serializeGt)

def _unwrapGt(x):
    return _unwrap(x, deserializeGt)

wrapX = _wrapG1
unwrapX = _unwrapG1
wrapY = _wrapGt
unwrapY = _unwrapGt




"""
Interface to the elliptic curve types and functions in the RELIC library.
"""
from relic import librelic
from bi import *
from ctypes import Structure, byref, sizeof, c_int, c_ubyte, c_ulonglong
from common import *

class ecElementBase(Structure):
    """
    Common base class for ec1, ec2, ec12 elements.
    """
    _elementType = "EC Element (Base)"
    _str_degree_max = None


    def __ne__(self, other):
        """
        Computes a != b as: not a == b
        """
        return not self.__eq__(other)


    def __rmul__(self, other):
        """
        Multiplies two elements since multiplication of EC points is 
        commutative.
        """
        return self.__mul__(other)


    def __str__(self, includeNormal=True):
        """
        Retrieves a string representation of this element with coordinates
        as hexadecimal strings.
        """
        if includeNormal:
            n = "Normalized" if self.normalized else "Not Normalized"
            name = self._elementType + " " + n
        else:
            name = self._elementType


        # Convert each point to a string, but limit the number of degrees
        # if necessary.
        d = self._str_degree_max if self._str_degree_max else self._degree

        formattedPoints = [formatPoint(p, self.normalized) for p in self.points[:d] ]
        pointText = "\n\n".join(formattedPoints)
        return name + "\n" + pointText


COORD_LEN = 4
class ecPoint(Structure):
    """
    Common base class for working with RELIC elliptic curve points.
    """
    # The size, in bytes, of our base field
    BASE_FIELD_BYTES = 256/8

    # The length of each coordinate (in bytes) for the type fp_t 
    # under our specified configuration.
    COORD_LEN = BASE_FIELD_BYTES/sizeof(c_ulonglong)

    _fields_ = [
                # x,y,z are of type fp_(s)t which are simple byte arrays of a 
                # fixed length.
                ("x", c_ulonglong*COORD_LEN), 
                ("y", c_ulonglong*COORD_LEN),
                ("z", c_ulonglong*COORD_LEN)
            ]


class ec1Element(ecElementBase):
    """
    Elliptic curve point struct type is the underlying structure of a G1
    element.
    """
    _elementType = "EC1 Element"
    _degree = 1

    # This maps to the type ep_t in the relic_ep.h. It is typedef'd to the 
    # g1_t type.
    _fields_ = [
            ("points", ecPoint*_degree),
            ("normalized", c_int)
        ]

    def __add__(self, other):
        """
        Adds to EC elements.
        """
        assertSameType(self, other)
        return relicResult(librelic.ec_add_abi, ec1Element, self, other)


    def __mul__(self, k):
        """
        Computes kP where P is this element and k is an integer type.
        """
        # Perform multiplication and return the result
        return relicResult(librelic.ec_mul_abi, ec1Element, self,
            coerceBigInt(k))


    def __eq__(self, other):
        """
        Compares this EC point against another or the identity element 0.
        """
        return _equal(self, other, 0, librelic.ec_cmp_abi)


    def isIdentity(self):
        """
        Determines if this point is the additive identity 0 or the point at 
        infinity.
        """
        return librelic.ec_is_infty_abi(byref(self)) == 1


    def normalize(self):
        """
        There is no in-place normalization for EC elements.
        """
        relicResult(librelic.ec_norm_abi, None, self, self)


class ec2Element(ecElementBase):
    """
    Element of a qudratic extension field over an elliptic curve.
    """
    _elementType = "EC2 Element"
    _degree = 2

    # This maps to the type ep2_(s)t in the relic_ep.h. It is typedef'd to the 
    # g2_t type.
    _fields_ = [
            ("points", ecPoint*_degree),
            ("normalized", c_int)
        ]


class ec12Element(ecElementBase):
    """
    Element of a dodectic extension field over an elliptic curve.
    """
    _elementType = "EC12 Element"
    _degree = 12

    # The number of degrees we will actually use when generating a string
    _str_degree_max = 4

    # DEBUG: Reduce the printing degrees so the output is readable
    _str_degree_max = 1

    # This maps to the type fp12_(s)t in the relic_ep.h. It is typedef'd to the 
    # gt_t type.
    _fields_ = [
            ("points", ecPoint*_degree),
            ("normalized", c_int)
        ]

class lwnafTable(Structure):
    """
    LWNAF table for EC2 precomputation.
    """
    # In our configuration, this is 4x the size of an ec2 element:
    # relic_epx.h: EPX_TABLE_LWNAF
    SIZE = 4
    _fields_ = [ ("values", ec2Element*SIZE) ]


def relicResult(relicFunc, resultType, *args):
    """
    Calls @relicFunc with a list of @args that are passed by reference. If
    @resultType is set, a new object of this type is created, passed as the
    first argument, and returned by this function.
    """
    result = None

    # If there is a return type, it becomes the first parameter.
    if resultType is not None:
        result = resultType()
        args = (result,) + args

    # Pass all parameters byref
    params = [byref(x) for x in list(args)]

    # Call the function and return our result.
    relicFunc(*params)
    return result


def _deserialize(x, elementType, compress, relicReadBinFunc):
    """
    Deserializes a bytearray @x, into an @element of the correct type,
    using the a relic read_bin function and the specified @compressed flag.
    This is the underlying implementation for deserialize G1, G2, and Gt.
    """
    # Convert the bytearray into an appropriately sized ctypes array of bytes
    b = (c_ubyte*len(x))(*bytearray(x))

    # The compression flag is an integer.
    flag = c_int(compress)

    # Deserialize using the read function.
    result = elementType()
    relicReadBinFunc(byref(result), byref(b), len(x), flag)
    return result


def _equal(a, b, identityLong, relicCompare):
    """
    Compares element @a to @b. If @b is @identityLong, returns 
    a.isIdentity(). Otherwise, normalize a and b and use the relicCompare
    function to test for equality.
    """
    # Check for an identity comparison.
    if isinstance(b, (long, int)) and b == identityLong:
        return a.isIdentity()

    # Verify type and fix normalization for a valid comparison.
    assertSameType(a, b)
    a.normalize()
    b.normalize()

    # Compare the elements using the relic function.
    return relicCompare(byref(a), byref(b)) == EQUAL


def _getCachedValue(obj, relicFunc, resultType):
    """
    Retrieves a value from obj.cached (if not None) or calls @relicFunc and 
    caches the result (of @resultType) int obj.cached.

    This is a common implementation for orderG1/G2/Gt and generatotG1/G2/Gt
    """
    # If the value has not been previously cached, fetch 
    if not obj.cached:
        obj.cached = resultType()
        relicFunc(byref(obj.cached))
    return obj.cached


def _serialize(element, compress, relicSizeBinFunc, relicWriteBinFunc):
    """
    Serializes an @element using the proper function @relicWriteBinFunc into
    a bytearray. @compress specifies whether the element should be compressed.
    @relicSizeBinFunc is used to determine the size of the serialized output.
    This is underlying implementation for serialize G1, G2, and Gt.
    """
    cFlag = c_int(compress)
    size = relicSizeBinFunc(byref(element), cFlag)

    # Make an array of the correct size. 
    binArray = (c_ubyte*size)()

    # Serialize
    relicWriteBinFunc(byref(binArray), size, byref(element), cFlag)
    return bytearray(binArray)


def deserializeEc(x, compress=True):
    """
    Deserialize binary string @x into an EC element.
    """
    return _deserialize(x, ec1Element, compress, librelic.ec_read_bin_abi)


def formatPoint(point, affine):
    """
    Retrieves a string representation of @point
    """
    # Affine coordinates: (x,y)        
    if affine:
        fmt = "\tx:{}\n\ty:{}"
        coords = [point.x, point.y]

    # Projected coordinates: (x,y,z)
    else:
        fmt = "\tx:{}\n\ty:{}\n\tz:{}"
        coords = [point.x, point.y, point.z]

    coordText = map(hexString, coords)
    return fmt.format(*coordText)


def generatorEc():
    """
    Retrieves the generator <G> = ECGroup
    """
    return _getCachedValue(generatorEc, librelic.ec_curve_get_gen_abi, 
        ec1Element)

generatorEc.cached = None


def randomEcPoint():
    """
    Generates a random element from the ECGroup.
    """
    return relicResult(librelic.ec_rand_abi, ec1Element)


def serializeEc(P, compress=True):
    """
    Generates a compact binary version of this point.
    """
    return _serialize(P, compress, librelic.ec_size_bin_abi,
        librelic.ec_write_bin_abi)


def orderEc():
    """
    Retrieves the order of the elliptic curve group as a BigInt.
    """
    return _getCachedValue(orderEc, librelic.ec_curve_get_ord_abi, BigInt)

orderEc.cached = None


# from vpopProfile import *
from vpop import *
from testcommon import *

# Arbitrary operands for our protocol
w = "Some super-secret ensemble key selector"
t = "Totally random and unpredictable tweak"
m = "This is a secret message"
msk = randomstr(n=64)
s = "Super secret table value"
hash_in = randomstr(n=64)

iterations = 10000

@profile
def primitives():
    """
    Perform primitive operations for profiling
    """
    z = randomZ(orderG1())

    # G1 operations
    P,Q = randomG1(),randomG1()
    R = generatorG1()
    g1Add = P + Q
    g1ScalarMultiply = z*P
    g1GeneratorMultiply = z*R
    g1Hash = hashG1(hash_in)

    # G2 operations
    P,Q = randomG2(),randomG2()
    R = generatorG2()
    g2Add = P + Q
    g2ScalarMultiply = z*P
    g2GeneratorMultiply = z*R
    g2hash = hashG2(hash_in)

    # Gt operations
    P = randomGt()
    Q = randomGt()
    gtMult = P * Q
    gtExp = P**z

    # Pairing
    x,y = (randomG1(), randomG2())
    R = pair(x,y)


@profile
def protoWithProof():
    """
    Run the full protocol including proof generation and verification.
    """
    r, x = blind(m)
    y,kw,tTilde = eval(w,t,x,msk,s)

    pi = prove(x, tTilde, kw, y)
    verify(x, t, y, pi, errorOnFail=True)

    z = deblind(r, y)


@profile
def proofMethods():
    """
    Run the full protocol including proof generation and verification.
    """
    r, x = blind(m)
    y,kw,tTilde = eval(w,t,x,msk,s)

    # Proof in Gt/Gt
    pi = proveGt(x, tTilde, kw, y)
    verifyGt(x, tTilde, y, pi, errorOnFail=True)

    # Proof in G1/Gt
    pi = proveG1(x, tTilde, kw, y)
    verifyG1(x, tTilde, y, pi, errorOnFail=True)

    z = deblind(r, y)


@profile
def protoProofGt():
    r, x = blind(m)
    y,kw,tTilde = eval(w,t,x,msk,s)
    pi = proveGt(x, tTilde, kw, y)
    verifyGt(x, tTilde, y, pi, errorOnFail=True)
    z = deblind(r, y)


@profile
def protoProofG1():
    r, x = blind(m)
    y,kw,tTilde = eval(w,t,x,msk,s)
    pi = proveG1(x, tTilde, kw, y)
    verifyG1(x, tTilde, y, pi, errorOnFail=True)
    z = deblind(r, y)


@profile
def proofMethods():
    """
    Run the full protocol including proof generation and verification.
    """
    r, x = blind(m)
    y,kw,tTilde = eval(w,t,x,msk,s)

    # Proof in Gt/Gt
    pi = proveGt(x, tTilde, kw, y)
    verifyGt(x, tTilde, y, pi, errorOnFail=True)

    # Proof in G1/Gt
    pi = proveG1(x, tTilde, kw, y)
    verifyG1(x, tTilde, y, pi, errorOnFail=True)

    z = deblind(r, y)


@profile   
def protoFast():
    """
    Runs the protocol but omits proof generation and verification.
    """
    r, x = blind(m)
    y,kw,tTilde = eval(w,t,x,msk,s)
    z = deblind(r, y)


def repeat(func, n=iterations):
    """
    Call @func @n times.
    """
    for _ in range(n):
        func()

# Run!
if __name__ == "__main__":
    repeat(primitives)
    # repeat(protoFast)
    # repeat(protoWithProof)
    # repeat(proofMethods)
    # repeat(protoProofG1)
    # repeat(protoProofGt)




"""
Interface to the Barreto Naehrig 256-bit pairing-based elliptic curves 
(PBC) in the RELIC library.
"""
from relic import librelic
from ctypes import byref, c_int, c_ubyte
from ec import *
from ec import _getCachedValue, _equal, _serialize, _deserialize
from bi import *
from common import *


class G1Element(ec1Element):
    """
    An element (member) of the additive group G1.
    """
    _elementType = "G1 Element"


    def __add__(self, other):
        """
        Adds two G1 elements to produce another G1 element.
        """
        return _add(self, other, librelic.g1_add_abi)        


    def __eq__(self, other):
        """
        Compares two G1 elements. Also determines if the point is the identity
        when calling "self == 0".
        """
        return _equal(self, other, 0, librelic.g1_cmp_abi)


    def __mul__(self, other):
        """
        Multiplies this G1Element with a scalar of integer type.
        """
        # Always prefer the generator multiply routine when possible. It's
        # roughly 2x faster.
        if self is generatorG1():
            return _genMultiply(other, G1Element, orderG1(), 
                librelic.g1_mul_gen_abi)

        # Otherwise use the normal scalar multiply routine.
        return _scalarMultiply(self, other, orderG1(), librelic.g1_mul_abi)


    def inverse(self):
        """
        Retrieves the inverse of a G1 element.
        """
        result = G1Element()
        librelic.g1_neg_abi(byref(result), byref(self))
        return result


    def isIdentity(self):
        """
        Determines if this element is the identity element of G1 (also called O
        or the point at infinity).
        """
        return librelic.g1_is_infty_abi(byref(self)) == 1


    def normalize(self):
        """
        Normalizes this element.
        """
        librelic.g1_norm_abi(byref(self), byref(self))


class G2Element(ec2Element):
    """
    An element (member) of the additive group G2.
    """
    _elementType = "G2 Element"

    # Precomputation table for multiplication.
    _table = None

    def __add__(self, other):
        """
        Adds two G2 elements to produce another G2 element.
        """
        return _add(self, other, librelic.g2_add_abi)


    def __eq__(self, other):
        """
        Compares two G1 elements. Also determines if the point is infinity
        (additive identity) when calling "self == 0".
        """
        return _equal(self, other, 0, librelic.g2_cmp_abi)


    def __mul__(self, other):
        """
        Multiplies this G2Element with a BigInt or Python long value.
        """
        # Always prefer the generator multiply routine when possible. It's
        # roughly 2x faster.
        if self is generatorG2():
            return _genMultiply(other, G2Element, orderG2(),
                librelic.g2_mul_gen_abi)

        # Multiplication in G2 is so slow. On our development platform
        # it was 33% faster to build and use a precomputation table using the 
        # LWNAF algorithm than to use the default, basic, multiplication.
        return self.mul_table(other)


    def inverse(self):
        """
        Retrieves the inverse of a G1 element.
        """
        result = G2Element()
        librelic.g2_neg_abi(byref(result), byref(self))
        return result


    def isIdentity(self):
        """
        Determines if this element is the additive identity element of G2 
        (also called O or the point at infinity).
        """
        return librelic.g2_is_infty_abi(byref(self)) == 1


    def mul_basic(self, other):
        """
        Multiplies this G2Element with a BigInt or Python long value using the
        basic RELIC multiplication algorithm.
        """
        return _scalarMultiply(self, other, orderG2(), librelic.g2_mul_abi)


    def mul_table(self, other):
        """
        Fast multiplication using a the LWNAF precomputation table.
        """
        # Get a BigInt
        other = coerceBigInt(other)
        if not other:
            return NotImplemented
        other %= orderG2()

        # Building the precomputation table, if there is not one already.
        if not self._table:
            self._table = lwnafTable()
            librelic.ep2_mul_pre_lwnaf(byref(self._table), byref(self))


        result = G2Element()
        librelic.ep2_mul_fix_lwnaf(byref(result), byref(self._table), 
            byref(other))
        return result


    def normalize(self):
        """
        Normalizes this element.
        """
        librelic.g2_norm_abi(byref(self), byref(self))



class GtElement(ec12Element):
    """
    An element (member) of the multiplicative group Gt.
    """
    _elementType = "Gt Element"


    def __eq__(self, other):
        """
        Compares two Gt elements. Also determines if the point is the unity
        (multiplicative identity) element of Gt when calling "self == 1".
        """
        return _equal(self, other, 1, librelic.gt_cmp_abi)


    def __invert__(self):
        """
        Computes the inverse of an element in Gt.
        """
        result = GtElement()
        librelic.gt_inv_abi(byref(result), byref(self))
        return result


    def __mul__(self, other):
        """
        Multiplies two GtElements.
        """
        assertSameType(self, other)
        result = GtElement()
        librelic.gt_mul_abi(byref(result), byref(self), byref(other))
        return result


    def __pow__(self, exp):
        """
        Computes self^exp where @exp is an integer type.
        """
        exp = coerceBigInt(exp)
        if not exp:
            return NotImplemented

        # Shrink large exponents.
        exp %= orderGt()

        r = GtElement()
        librelic.gt_exp_abi(byref(r), byref(self), byref(exp))
        return r


    def normalize(self):
        """
        GtElements don't support normalization (apparently) in RELIC.
        """
        pass


    def isIdentity(self):
        """
        Determines if this element is the unit (multiplicative identity) of Gt.
        """
        return self.isUnity


    def isUnity(self):
        """
        Determines if this element is the unit (multiplicative identity) of Gt.
        """
        return librelic.gt_is_unity_abi(byref(self)) == 1


def _add(a, b, relicAdd):
    """
    Adds two elements @a,@b of the same type into @result using @relicAddFunc.
    """
    # Check types, create a result object of the same type, and call the relic
    # function.
    assertSameType(a,b)
    result = type(a)()
    relicAdd(byref(result), byref(a), byref(b))
    return result



def _scalarMultiply(P, a, n, relicScalarMult):
    """
    Performs scalar multiplication between point P \in G, scalar a \in Z, 
    using the function @relicScalarMult. @n is the order of the group G.
    """
    # Ensure the scalar is a BigInt
    a = coerceBigInt(a)
    if not a:
        return NotImplemented

    # Shrink large scalars.
    a %= n

    # Create a point to hold the result and multiply.
    result = type(P)()
    relicScalarMult(byref(result), byref(P), byref(a))
    return result


def _genMultiply(a, element, n, relicGenMultiplyFunc):
    """
    Multiplies scalar @a by the group generator using @relicGenMultiplyFunc
    and returns the result of type @element.
    """
    # Ensure the scalar is a BigInt
    a = coerceBigInt(a)
    if not a:
        return NotImplemented

    # Shrink large scalars.
    a %= n

    result = element()
    relicGenMultiplyFunc(byref(result), byref(a))
    return result


def deserializeG1(x, compressed=True):
    """
    Deserializes an array of bytes, @x, into a G1 element.
    """
    return _deserialize(x, G1Element, compressed, librelic.g1_read_bin_abi)


def deserializeG2(x, compressed=True):
    """
    Deserializes an array of bytes, @x, into a G2 element.
    """
    return _deserialize(x, G2Element, compressed, librelic.g2_read_bin_abi)


def deserializeGt(x, compressed=True):
    """
    Deserializes an array of bytes, @x, into a Gt element.
    """
    return _deserialize(x, GtElement, compressed, librelic.gt_read_bin_abi)


def generatorG1():
    """
    Retrieves the generator <P> = G1
    """
    return _getCachedValue(generatorG1, librelic.g1_get_gen_abi, G1Element)


def generatorG2():
    """
    Retrieves the generator <P> = G2
    """
    return _getCachedValue(generatorG2, librelic.g2_get_gen_abi, G2Element)


def generatorGt():
    """
    Retrieves the generator <g> = Gt
    """
    return _getCachedValue(generatorGt, librelic.gt_get_gen, GtElement)


# Initialize generator cached values to None
generatorG1.cached, generatorG2.cached, generatorGt.cached = None, None, None


def getBuffer(x):
    """
    Copy @x into a (modifiable) ctypes byte array
    """
    b = bytes(x)
    return (c_ubyte * len(b)).from_buffer_copy(bytes(x))



def _hash(x, elementType, relicHashFunc):
    """
    Hash an array of bytes, @x, using @relicHashFunc and returns the result
    of @elementType.
    """
    # Combine all inputs into a single bytearray
    barray = bytearray()
    map(barray.extend, bytes(x))

    # Create an element of the correct type to hold the hash result
    result = elementType()

    # Convert barray into a modifiable ctypes buffer, hash using the provided
    # function, and return the result.
    buf = getBuffer(barray)
    relicHashFunc(byref(result), byref(buf), sizeof(buf))
    return result


def hashG1(*args):
    """
    Hash an array of bytes, @x, onto the group G1. 
    @returns a G1Element.
    """
    return _hash(args, G1Element, librelic.g1_map_abi)


def hashG2(*x):
    """
    Hash an array of bytes, @x, onto the group G2.
    @returns a G2Element.
    """ 
    return _hash(x, G2Element, librelic.g2_map_abi)


def orderG1():
    """
    Retrieves the order (size) of group G1 as a BigInt.
    """
    return _getCachedValue(orderG1, librelic.g1_get_ord_abi, BigInt)


def orderG2():
    """
    Retrieves the order (size) of group G2 as a BigInt.
    """
    return _getCachedValue(orderG2, librelic.g2_get_ord_abi, BigInt)


def orderGt():
    """
    Retrieves the order (size) of group Gt as a BigInt.
    """
    return _getCachedValue(orderGt, librelic.gt_get_ord_abi, BigInt)


# Initialize cached values to None
orderG1.cached, orderG2.cached, orderGt.cached = None, None, None


def pair(p,q):
    """
    Computes the bilinear pairing e(p,q). @p must be a G1Element and @q must
    be a G2Element.
    @returns a GtElement
    """
    # Check types
    assertType(p, G1Element)
    assertType(q, G2Element)

    result = GtElement()
    librelic.pc_map_abi(byref(result), byref(p), byref(q))
    return result


def _random(elementType, relicRandomFunc):
    """
    Retrieves a random element of @elementType by calling @relicRandomFunc.
    """
    result = elementType()
    relicRandomFunc(byref(result))
    return result


def randomG1():
    """
    Select a random element from G1.
    """
    return _random(G1Element, librelic.g1_rand_abi)


def randomG2():
    """
    Select a random element from G2.
    """
    return _random(G2Element, librelic.g2_rand_abi)


def randomGt():
    """
    Select a random element from Gt.
    """
    return _random(GtElement, librelic.gt_rand)


def serializeG1(x, compress=True):
    """
    Converts G1 element @x into an array of bytes. If @compress is True, 
    the point will be compressed resulting in a much shorter string of bytes.
    """
    assertType(x, G1Element)
    return _serialize(x, compress, librelic.g1_size_bin_abi,
        librelic.g1_write_bin_abi)


def serializeG2(x, compress=True):
    """
    Converts G2 element @x into an array of bytes. If @compress is True, 
    the point will be compressed resulting in a much shorter string of bytes.
    """
    assertType(x, G2Element)
    return _serialize(x, compress, librelic.g2_size_bin_abi,
        librelic.g2_write_bin_abi)


def serializeGt(x, compress=True):
    """
    Converts Gt element @x into an array of bytes. If @compress is True, 
    the point will be compressed resulting in a much shorter string of bytes.
    """
    assertType(x, GtElement)
    return _serialize(x, compress, librelic.gt_size_bin_abi,
        librelic.gt_write_bin_abi)




#!/usr/bin/eval python

"""
Tests for arithmetic for G1, G2, and Gt elements.
"""

from testcommon import *
from pbc import *
from timeit import timeit
from unittest import TestCase, SkipTest
import unittest


class AdditiveGroupArithmetic(TestCase):
    """
    Tests G1/G2 arithmetic
    """
    def setUp(self):
        raise SkipTest("base class")


    def additionCommutes(self):
        P, Q = self.randomElement(), self.randomElement()
        R1 = P + Q
        R2 = Q + P
        self.assertNotEqual(P, Q)
        self.assertNotEqual(R1, Q)
        self.assertEqual(R1, R2)


    def additionAssociates(self):
        P, Q, R = self.randomElement(), self.randomElement(), self.randomElement()
        T1 = (P + Q) + R
        T2 = P + (Q + R)
        self.assertEqual(T1, T2)


    def additionDistributes(self):
        Q = self.randomElement()
        a, b = randomZ(), randomZ()
        c = a + b

        R1 = Q*a + Q*b
        R2 = Q*c

        self.assertNotEqual(Q, R1)
        self.assertEqual(R1, R2)


    def largeScalarMultiply(self):
        n = self.order
        Q = self.randomElement()
        a,b = randomZ(maximum=n), randomZ(maximum=n)
        c = a * b

        # Ensure our number is big enough for this test
        self.assertTrue(c > n)

        # Perform multiplication with two smaller numbers and one larger
        # number. They should give the same result and no segfaults.
        R1 = Q*a*b
        R2 = Q*c
        self.assertEqual(R1, R2)


    def testAdditionCommutes(self):
        repeat(self.additionCommutes, n=100)

    def testAdditionAssociates(self):
        repeat(self.additionAssociates, n=100)

    def testAdditionDistributive(self):
        repeat(self.additionDistributes, n=100)


    def testMultiplyReduces(self):
        Q = self.randomElement()
        a,b = randomZ(), randomZ()
        Qa = Q*a
        Qab = Qa*b
        c = a*b % self.order
        Qc = Q*c
        self.assertEqual(Qc, Qab)


    def testLargeScalarMultiply(self):
        repeat(self.largeScalarMultiply, n=1)


    def testMultiplyCommutes(self):
        """
        Test multiplication by testing commutativity.
        """
        g = self.randomElement()
        a = randomZ()
        b = randomZ()
        c = a*b

        h1 = (g*a)*b
        h2 = (g*b)*a
        h3 = g*c
        self.assertNotEqual(g, h1)
        self.assertEqual(h1, h2)
        self.assertEqual(h2, h3)


    def testInversion(self, n=100):
        """
        Tests G1 element inversion by multiplying computing inverses and 
        multiplying.
        """
        def randomInv():
            g = self.randomElement()
            gInv = g.inverse()
            h = g + gInv
            self.assertTrue(h == 0)
            self.assertTrue(h+g == g)
        repeat(randomInv, n)


class G1Tests(AdditiveGroupArithmetic):
    """
    Tests for G1Element arithmetic.
    """
    def setUp(self):
        self.randomElement = randomG1
        self.order = orderG1()


class G2Tests(AdditiveGroupArithmetic):
    """
    Tests for G2Element arithmetic.
    """
    def setUp(self):
        self.randomElement = randomG2
        self.order = orderG2()

    def testFastMultiplyG2Correct(self):
        """
        Tests that fast multiplication is correct by cross-checking with slow 
        multiplication.
        """
        p = randomG2()
        r = randomZ()
        q1 = p.mul_table(r)
        q2 = p.mul_basic(r)
        self.assertEqual(q1, q2)


    def testFastMultiplyG2Faster(self):
        """
        Ensures that fast multiplication is indeed faster than basic multiply.
        """
        p = randomG2()
        r = randomZ()

        fastTime = timeit(lambda:p.mul_table(r), number=100)
        basicTime = timeit(lambda:p.mul_basic(r), number=100)
        self.assertLess(fastTime, basicTime)


class GtTests(TestCase):
    """
    Tests for GtElement arithmetic
    """
    def testInverse(self):
        """
        Verifies that ~GtElement produces a multiplicative inverse.
        """
        def doInv():
            r = randomGt()
            rInv = ~r
            x = r * rInv
            self.assertTrue(x == 1)
        repeat(doInv, n=50)


    def testMultiplyCommutes(self):
        def test():    
            g, h = randomGt(), randomGt()
            r1 = g*h
            r2 = h*g
            self.assertNotEqual(g, r1)
            self.assertEqual(r1,r2)

        repeat(test, n=100)


    def testMultiplyAssociates(self):
        def test():
            f, g, h = randomGt(), randomGt(), randomGt()
            r1 = (f*g)*h
            r2 = f*(g*h)
            self.assertEqual(r1,r2)

        repeat(test, n=100)


    def testExpCommutes(self):
        def test():
            g = randomGt()
            a,b = randomZ(), randomZ()
            r1 = (g**a)**b
            r2 = (g**b)**a
            r3 = g**(a*b)
            c = (a*b) % orderGt()
            r4 = g**c
            self.assertEqual(r1, r2)
            self.assertEqual(r1, r3)
            self.assertEqual(r1, r4)

        repeat(test, n=100)


    def testExpDistributes(self):
        def test():
            g = randomGt()
            a,b = randomZ(), randomZ()
            r1 = (g**a)*(g**b)
            r2 = g**(a+b)
            self.assertEqual(r1, r2)

        repeat(test, n=100)


# Run!
if __name__ == '__main__':
    unittest.main()


"""
An (unblinded) verifiable, pseudorandom function (V PRF) designed to work with 
Pythia API. For simplicity, this particular implementation uses G1 from the BN
pairing-based curves, but any elliptic curve with sufficient security would 
also suffice.
"""
from pbc import *
from common import *
from prf import *

def eval(w,t,x,msk,s):
    """
    Pythia server-side computation of intermediate PRF output.
    @w: ensemble key selector (any string, e.g. webserver ID)
    @t: tweak (any string, e.g. user ID)
    @x: message (any string)
    @msk: Pythia server's master secret key
    @s: state value from Pythia server's key table
    @returns: (y, kw, dummy=None)
     where: y: intermediate result
            kw: secret key bound to w (needed for proof)
            beta: H(kw,t,x) (needed for proof)
    """
    # Verify types
    assertType(w, (str, int, long))
    assertType(t, (str, int, long))
    assertType(x, (str, int, long))

    # Construct the key
    kw = genKw(w,msk,s)

    # Compute y
    beta = hashG1(t, x)
    y = beta*kw
    return y,kw,beta


def prove(x,beta,kw,y):
    """
    Generate a zero-knowledge proof that DL(Q*kw) = DL(beta*kw) where
    <Q> = G1.
    """
    # Verify types
    assertScalarType(kw)
    assertType(beta, G1Element)
    assertType(y, G1Element)

    # Compute the proof.
    Q = generatorG1()
    p = Q*kw
    v = randomZ(orderG1())
    t1 = Q*v
    t2 = beta*v

    t1.normalize()
    t2.normalize()

    c = hashZ(Q,p,beta,y,t1,t2)
    u = (v-(c*kw)) % orderG1()
    return (p,c,u)


def verify(x, t, y, pi, errorOnFail=True):
    """
    Verifies a zero-knowledge proof.
    @errorOnFail: Raise an exception if the proof does not hold.
    """
    # Unpack the proof
    p,c,u = pi

    # Verify types
    assertType(x, str)
    assertType(t, str)
    assertType(y, G1Element)
    assertType(p, G1Element)
    assertScalarType(c)
    assertScalarType(u)

    # TODO: beta can be pre-computed while waiting for a server response.
    Q = generatorG1()
    beta = hashG1(t, x)

    # Recompute c'
    t1 = Q*u + p*c 
    t2 = beta*u + y*c

    t1.normalize()
    t2.normalize()

    cPrime = hashZ(Q,p,beta,y,t1,t2)

    # Check computed @c' against server's value @c
    if cPrime == c:
        return True

    if errorOnFail:
        raise Exception("zero-knowledge proof failed verification.")
    else:
        return False


# Blind/deblind are more or less the identity function. Included for
# API compatibility with the other Pythia PRFs.
blind = lambda m: (None, m)

def deblind(r,x):
    return x

# unwrap (decode and deserialize) elements by name
unwrapY = unwrapG1
unwrapP = unwrapG1
unwrapC = unwrapLong
unwrapU = unwrapLong

# We don't do any decoding or deserialization of the message parameter x in
# this version of the protocol.
unwrapX = lambda x: x



#!/usr/bin/eval python
from testcommon import *
import unittest
from unittest import TestCase
from bls import *
from common import hmac

class BlsTests(TestCase):
    """
    Tests for the BLS PRF
    """
    def testEvalStable(self, n=10):
        """
        Runs eval() @n times and ensures that after deblinding, the result
        is deterministic.
        """
        Y = None
        w = "Some super-secret ensemble key selector"
        t = "Totally random and unpredictable tweak"
        m = "This is a secret message"
        msk = "lkjasdf;lkjas;dlkfa;slkdf;laskdjf"
        s = "Super secret table value"

        for _ in range(n):
            y,_,_ = eval(w,t,m,msk,s)

            # Save the first output
            if not Y:
                Y = y

            # Compare each output to the first.
            self.assertTrue(y == Y)


    def testProof(self):
        """
        Tests that the zero-knowledge proof passes verification when generated 
        using randomly selected inputs.
        """
        kw = randomZ()
        m = randomstr()
        t = randomstr()
        beta = hashG1(t, m)
        y = beta*kw

        pi = prove(None, None, kw, None)
        self.assertTrue( verify(m, t, y, pi, errorOnFail=False) )


    def testBadProof(self):
        """
        Tests that an invalid proof is reported as invalid.
        """
        kw = randomZ()
        m = randomstr()
        t = randomstr()
        y = randomG1()

        pi = prove(None, None, kw, None)
        self.assertFalse( verify(m, t, y, pi, errorOnFail=False) )


    def testBadPubkey(self):
        """
        Tests that an invalid proof with a bad pubkey is reported invalid.
        """
        # Generate a correct result @y and proof using random inputs.
        kw = randomZ()
        m = randomstr()
        t = randomstr()
        beta = hashG1(t, m)
        y = beta*kw

        # Generate a valid proof
        (p,c,u) = prove(None, None, kw, None)

        # Swap out the pubkey p with a bogus value
        badP = randomG2()
        pi = (badP, None, None)

        self.assertFalse( verify(m, t, y, pi, errorOnFail=False) )


    def testVerifyFailRandom(self, n=100):
        """
        Tests that random inputs are not reported as a valid proof.
        """
        m = randomstr()
        t = randomstr()
        y = randomG1()
        pi = (randomG2(), None, None)
        self.assertFalse( verify(m, t, y, pi, errorOnFail=False) )


    def testProtocol(self):
        """
        Tests a full pass of the protocol.
        """
        w = "Some super-secret ensemble key selector"
        t = "Totally random and unpredictable tweak"
        m = "This is a secret message"
        salt = randomstr()
        x = hmac(salt, m)
        msk = randomstr()
        s = randomstr()

        # Run the protocol 
        y,kw,beta = eval(w,t,x,msk,s)
        pi = prove(None, None, kw, None)

        # Check the proof
        self.assertTrue( verify(x, t, y, pi, errorOnFail=False) )


# Run!
if __name__ == '__main__':
    unittest.main()


"""
Common routines for Pythia pseudorandom function (PRF) protocols based on 
pairing based curves (BN-254).
"""
from pbc import *
import base64

def genKw(w,msk,z):
    """
    Generates key Kw using key-selector @w, master secret key @msk, and
    table value @z.
    @returns Kw as a BigInt.
    """
    # Hash inputs into a string of bytes
    b = hmac(msk, z + w, tag="TAG_PYTHIA_KW")

    # Convert the string into a long value (no larger than the order of Gt),
    # then return a BigInt value.
    return BigInt(longFromString(b) % long(orderGt()))


def getDelta(original, update):
    """
    Generates an update token delta_{k->k'}.
    @original: values for k: (w,msk,s)
    @update: values for kPrime: (w',msk',s')
    @return (delta, p'): @delta = k'/k, @p' is a new pubkey based on k'.
    """
    # Compute both keys
    k = genKw(*original)
    kPrime = genKw(*update)

    # Compute delta,p'
    delta = (kPrime * inverse(k, orderGt())) % orderGt()
    pPrime = generatorGt()**kPrime
    return delta,pPrime


def update(z,delta):
    """
    Updates a result @z using the update token @delta.
    """
    return z**delta

    
def wrap(x):
    """
    Wraps an element or integer type by serializing it and base64 encoding 
    the resulting bytes.
    """
    # Detect the type so we can call the proper serialization routine
    if isinstance(x, G1Element):
        return _wrap(x, serializeG1)

    elif isinstance(x, G2Element):
        return _wrap(x, serializeG2)

    elif isinstance(x, GtElement):
        return _wrap(x, serializeGt)

    elif isinstance(x, str):        
        return x

    elif isinstance(x, (int, long, BigInt)):
        return hex(long(x))

    # All other items
    else:
        raise NotImplementedError("Cannot unwrap {}; only types {} supported".
            format(type(x), 
                [G1Element, G2Element, GtElement, int, long, BigInt]) )


# Individual unwrap functions
unwrapStr = lambda x: x
unwrapG1 = lambda x: _unwrap(x, deserializeG1)
unwrapG2 = lambda x: _unwrap(x, deserializeG2)
unwrapGt = lambda x: _unwrap(x, deserializeGt)
unwrapLong = lambda x: long(x, 16)
unwrapDelta = unwrapLong


def _wrap(x, serializeFunc, encodeFunc=base64.urlsafe_b64encode, compress=True):
    """
    Wraps an element @x by serializing and then encoding the resulting bytes.
    """
    return encodeFunc(serializeFunc(x, compress))


def _unwrap(x, deserializeFunc, decodeFunc=base64.urlsafe_b64decode, compress=True):
    """
    Unwraps an element @x by decoding and then deserializing
    """
    return deserializeFunc(decodeFunc(str(x)), compress)




#!/usr/bin/eval python

from testcommon import *
import unittest
from unittest import TestCase
from vpop import *

# Global values for test cases
w = "Some super-secret ensemble key selector"
t = "Totally random and unpredictable tweak"
m = "This is a secret message"
msk = "lkjasdf;lkjas;dlkfa;slkdf;laskdjf"
s = "Super secret table value"


class VpopTests(TestCase):
    """
    Tests for the Vpop class.
    """
    def testKeyGen(self):
        msk = "12l3k4jh1lk23jh51l34b5l1k34j5nl134jn51lk3b51lk34b5"
        w = "super secret key selector"
        z = "lkjasdf;lkjaew;jas; dlkjas;ldfkjas;dlfkjas ;df"
        genKw(w,msk,z)


    def testBlindRandomized(self, n=100):
        """
        Verifies that the blinding method is randomized.
        """
        m = "This is a very sensitive passphrase."
        def doBlind():
            _, x = blind(m)
            return x

        randomNoRepeat(doBlind, n=n)


    def testEvalStable(self, n=100):
        """
        Runs eval() @n times and ensures that after deblinding, the result
        is deterministic.
        """
        Z = None

        for _ in range(n):
            r, x = blind(m)
            y,_,_ = eval(w,t,x,msk,s)
            z = deblind(r, y)

            # Save the first output
            if not Z:
                Z = z 

            # Compare each output to the first.
            self.assertTrue(z == Z)


    def testProof(self):
        """
        Tests that the zero-knowledge proof passes verification when generated 
        using randomly selected inputs.
        """
        # TODO: Consolidate random selection to a test setup to speed up unit
        # tests
        kw = randomZ()
        x = randomG1()
        t = ";lkjasd;flkqj23;lrkqm2d;lkmq3;klmq3; tcq93u4 t[0q34 pq9j43p9jq3 4p"
        tTilde = hashG2(t)
        y = pair(x*kw, tTilde)

        pi = prove(x, tTilde, kw, y)
        self.assertTrue( verify(x, t, y, pi, errorOnFail=False) )


    def testBadProof(self):
        """
        Tests that an invalid proof is reported as invalid.
        """
        kw = randomZ()
        x = randomG1()
        t = long(randomZ())
        tTilde = hashG2(t)
        y = randomGt()

        pi = prove(x, tTilde, kw, y)
        self.assertFalse( verify(x, t, y, pi, errorOnFail=False) )


    def testBadPubkey(self):
        """
        Tests that an invalid proof with a bad pubkey is reported invalid.
        """
        kw = randomZ()
        x = randomG1()
        t = randomZ()
        tTilde = hashG2(t)
        y = pair(x*kw, tTilde)

        # Generate a valid proof
        (p,c,u) = prove(x, tTilde, kw, y)

        # Swap out the pubkey p with a bogus value
        badP = randomG1()
        pi = (badP, c, u)

        self.assertFalse( verify(x, t, y, pi, errorOnFail=False) )


    def testVerifyFailRandom(self, n=100):
        """
        Tests that random inputs are not reported as a valid proof.
        """
        x = randomG1()
        t = randomZ()
        y = randomGt()
        pi = (randomG1(), randomZ(orderGt()), randomZ(orderGt()))
        self.assertFalse( verify(x, t, y, pi, errorOnFail=False) )


    def testProtocol(self):
        """
        Tests a full pass of the protocol.
        """
        # Run the protocol 
        r, x = blind(m)
        y,kw,tTilde = eval(w,t,x,msk,s)
        pi = prove(x, tTilde, kw, y)
        z = deblind(r, y)

        # Check the proof
        self.assertTrue( verify(x, t, y, pi) )


    def testProtocolStable(self, n=10):
        """
        Tests that multiple passes of the protocol produce the same z,p values.
        """
        # Run the protocol and report the z,p values
        def proto():
            r, x = blind(m)
            y,kw,tTilde = eval(w,t,x,msk,s)
            (p,_,_) = prove(x, tTilde, kw, y)
            z = deblind(r, y)
            return z,p

        # Establish initial values
        Z,P = proto()

        # Run the protocol and check the results against expected results
        def protoCheck():
            z,p = proto()
            self.assertTrue(z == Z)
            self.assertEqual(P, p)

        repeat(protoCheck, n=n)


    def testUpdate(self):
        """
        Tests updating an encrypted value to a new key.
        """
        wPrime = "Definitely not the original value w"
        sPrime = "Totally new state value"

        # Run the protocol under the original inputs
        z = simpleProto(w,t,msk,s,m)

        # Get an update token and apply it
        delta,_ = getDelta((w,msk,s), (wPrime,msk,sPrime))
        zPrime1 = update(z, delta)

        # Re-run the protocol under the new inputs
        zPrime2 = simpleProto(wPrime,t,msk,sPrime,m)

        # Compare z' 
        self.assertEqual(zPrime1, zPrime2)


def simpleProto(w,t,msk,s,m):
    """
    Runs the protocol without generating or checking proofs and returns
    the result @z.
    """
    r,x = blind(m)
    y,_,_ = eval(w,t,x,msk,s)
    return deblind(r, y)


# Run!
if __name__ == '__main__':
    unittest.main()


#!/usr/bin/eval python

from testcommon import *
import unittest, random
from bi import *
from relic import *


class BigIntTests(unittest.TestCase):
    """
    Tests for the BigInt class and related module functions.
    """
    def testFromInt(self):
        """
        Test creating BigInts from Python int values.
        """
        for i in range(1000):
            b = BigInt(i)
            self.assertEqual(i, long(b))


    def testEquality(self):
        a = random.randrange(0, 2*256)
        b = BigInt(a)
        c = BigInt(a)

        self.assertEqual(b, c)
        self.assertTrue(b == c )

        self.assertEqual(c, b)
        self.assertTrue(c == b )

        self.assertFalse(b != c )
        self.assertFalse(c != b )

        self.assertTrue( b >= c )
        self.assertTrue( b <= c )

        self.assertTrue( c >= b )
        self.assertTrue( c <= b )

        self.assertFalse( b < c )
        self.assertFalse( b > c )


    def testGtLtNe(self):
        a = randomZ()
        b = a+a

        self.assertNotEqual(a, b)
        self.assertTrue( a != b )

        self.assertFalse( b == a )
        self.assertFalse( a == b )

        self.assertTrue( b > a )
        self.assertTrue( b >= a )

        self.assertFalse( b < a )
        self.assertFalse( b <= a )


    def testFromLong(self, bits=192):
        """
        Test creating BigInts from Python long values by going full circle.
        """
        def randomConv():
            i = random.getrandbits(bits)
            b = BigInt(i)
            self.assertEqual(i, long(b))
        repeat(randomConv)


    def testLongMult(self, n=1000):
        def randomMult():
            # 2 random BigInts
            a, b = (randomZ(bits=256), randomZ(bits=256))
            c = a*b

            # Repeat with Python longs and check
            x, y = (long(a), long(b))
            z = x * y 

            self.assertEqual(z, long(c)) 
        repeat(randomMult, n)

    def testRandom8bit(self):
        self._testRandomMax(8)

    def testRandom32bit(self):
        self._testRandomMax(32)

    def testRandom64bit(self):
        self._testRandomMax(64)

    def testRandom256bit(self):
        self._testRandomMax(256)

    def testRandom1k(self):
        self._testRandomMax(1024)

    def _testRandomMax(self, bits, iterations=1000):
        # Grab a series of random values and ensure they're not too large.
        maxValue = 2**bits - 1
        for _ in range(iterations):
            b = randomZ(bits=bits)
            l = long(b)
            self.assertTrue( l <= maxValue )


# Run!
if __name__ == '__main__':
    unittest.main()


"""
Elliptic curve Qu-Vanstone implicit certificate scheme.
"""
from relic import librelic
from ec import *
from bi import *
from common import *


def _exp(cert, idText):
    """
    Generates the exponent e by hashing @cert and @idText.
    """
    return hashZ(serializeEc(cert) + idText)


def sign(idText, request, caPrivkey):
    """
    A certificate authority (CA) generates an implicit certificate using 
    identity string @id, @request (certificate public key component), and 
    the CA's private key @caPrivkey.
    @returns (s, cert) where @r is the private key contribution and @cert is 
     the implicit certificate.
    """
    # Verify input types
    assertType(request, ec1Element)
    assertScalarType(caPrivkey)

    # Switch to new notation
    R = request
    d = caPrivkey
    G = generatorEc()
    N = orderEc()

    # TODO: use simultaneous mul to speed this up.
    # Random integer
    k = randomZ(N)
    P = R + k*G

    # Hash the identity string and implicit cert into an integer
    e = _exp(P, idText)

    # Compute the private key contribution
    r = (e*k + d) % N
    return (r, P)


def validate(idText, alpha, r, cert, caPubkey):
    """
    A server can validate an implicit certificate response using identity
    string @idText, private value @alpha (used to generate cert request),
    and the certificate response @r (private key component) and implicit
    @cert.
    @raises Exception if the certificate response is invalid.
    @returns (privkey, pubkey)
    """
    # Verify parameter types
    assertScalarType(alpha)
    assertScalarType(r)
    assertType(cert, ec1Element)
    assertType(caPubkey, ec1Element)

    G = generatorEc()

    # Compute the private key @s
    e = _exp(cert, idText)
    s = (e*alpha + r) % orderEc()

    # Compute the public key two ways: using the privkey and using the cert
    # (the way a client will compute it)
    # The easy way
    S1 = s*G

    # Using the cert
    S2 = e*cert + caPubkey

    # The two techniques should produce the same pubkey value -- raise an
    # exception if they don't match
    if S1 != S2:
        raise Exception("Implicit certification response failed validation")
    return s, S1


def recoverPubkey(idText, cert, caPubkey):
    """
    A client can recover the server's pubkey using the identity string @idText,
    server's implicit @cert, and the trusted @caPubkey.
    """
    # Verify types
    assertType(cert, ec1Element)
    assertType(caPubkey, ec1Element)

    # Compute the pubkey
    return _exp(cert, idText)*cert + caPubkey



"""
Common routines
"""    
import binascii, hashlib
import hmac as HMAC

# RELIC comparison flags from relic_core.h
LESS_THAN = -1
EQUAL = 0
GREATER_THAN = 1
NOT_EQUAL = 2


def assertSameType(a, b):
    """
    Raises an exception if @b is not an instance of type(@a)
    """
    if not isinstance(b, type(a)):
        raise NotImplementedError("This operation is only supported for " \
            "elements of the same type. Instead found {} and {}".
                format(type(a), type(b)))


def assertType(var, *allowedTypes):
    """
    Asserts that a variable @var is of an @expectedType. Raises a TypeError
    if the assertion fails.
    """
    if not isinstance(var, *allowedTypes):
        raise NotImplementedError("This operation is only supported for {}. "\
            "Instead found {}".format(str(*allowedTypes), type(var)))


def dp(**kwargs):
    """
    Debugging print. Prints a list of labels and values, each on their
    own line.
    """
    for label,value in kwargs.iteritems():
            print "{0}\t{1}".format(label, value)


def hmac(key, message, tag=None, alg=hashlib.sha256):
    """
    Generates a hashed message authentication code (HMAC) by prepending the
    specified @tag string to a @message, then hashing with to HMAC 
    using a cryptographic @key and hashing @alg -orithm.
    """
    return HMAC.new(str(key), str(tag) + str(message), digestmod=alg).digest()


def hexString(values):
    """
    Generate a readable hexadecimal string from a list of binary @values.
    """
    # NOTE: abs(v) is used to discard leading minus sign (-)
    return " ".join([format(abs(v), 'X') for v in values])


def longFromString(x):
    """
    Convert arbitrary strings to long values.
    """
    return int(binascii.hexlify(x), 16)





"""
PIP setup script for the pyrelic package.
"""
from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

description=\
  """pyrelic is a python module to interact with the RELIC cryptography library
  which performs high-performance elliptic curve cryptography and pairing-based
  elliptic curves. pyrelic also includes verifiable pseudorandom functions
  used in Pythia.
  """
description = ' '.join(description.split())

setup(name='pythia-pyrelic',
      version='1.0.1',
      description=description,
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Operating System :: MacOS',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX :: Linux',
          'Operating System :: Unix',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Topic :: Security', 
          'Topic :: Security :: Cryptography',
      ],
      url='https://github.com/ace0/pyrelic',
      author='Adam Everspaugh',
      author_email='ace@cs.wisc.edu',
      license='MIT',
      keywords='encryption, elliptic curve cryptography, pairing based cryptography, oblvious pseudorandom function, partially oblivious pseudorandom function',
      package_data= { 'pyrelic': ['lib/*'] },
      packages=['pyrelic'],
      zip_safe=False, 
    )



"""
An (unblinded) pseudorandom function based on Boneh-Lynn-Shacham signatures 
designed for the Pythia API using BN254 pairing based curves.
"""
from pbc import *
from common import *
from prf import *


def eval(w,t,x,msk,s):
    """
    Pythia server-side computation of intermediate PRF output.
    @w: ensemble key selector (any string, e.g. webserver ID)
    @t: tweak (any string, e.g. user ID)
    @x: message (any string)
    @msk: Pythia server's master secret key
    @s: state value from Pythia server's key table
    @returns: (y, kw, dummy=None)
     where: y: intermediate result
            kw: secret key bound to w (needed for proof)
            dummy: None (included only for compatibility with Pythia PRF API)
    """
    # Verify types
    assertType(w, (str, int, long))
    assertType(t, (str, int, long))
    assertType(x, (str, int, long))

    # Construct the key
    kw = genKw(w,msk,s)

    # Compute y
    y = hashG1(t, x)*kw
    return y,kw,None


def prove(x,t,kw,y):
    """
    Computes public key P*kw where <P> = G1. 
    x, t, and y are ignored. They are included only for API compatibility with 
    other Pythia PRF implementations.
    """
    # Verify the key type and compute the pubkey
    assertScalarType(kw)
    p = generatorG2() * kw
    return (p,None,None)


def verify(x, t, y, pi, errorOnFail=True):
    """
    Verifies a zero-knowledge proof.
    @errorOnFail: Raise an exception if the proof does not hold.
    """
    # Unpack the proof
    p,_,_ = pi

    # Verify types
    assertType(x, str)
    assertType(t, str)
    assertType(y, G1Element)
    assertType(p, G2Element)

    # TODO: beta can be pre-computed while waiting for a server response.
    beta = hashG1(t, x)

    # Compute q = e( H(t,m), P)**kw two ways
    q1 = pair(beta, p)
    q2 = pair(y, generatorG2())

    # The BLS signature is valid when q1 == q2
    if q1 == q2:
        return True

    if errorOnFail:
        raise Exception("BLS signature failed verification")
    else:
        return False


# Blind/deblind are more or less the identity function. Only included for
# API compatibility with the other Pythia PRFs.
blind = lambda m: (None, m)

def deblind (r,x): 
    return x

# unwrap (decode and deserialize) elements by name
unwrapY = unwrapG1
unwrapP = unwrapG2
unwrapC = lambda x: None
unwrapU = lambda x: None

# We don't do any decoding or deserialization of the message parameter x in
# this version of the protocol.
unwrapX = lambda x: x



"""
Access to the RELIC multiple precision integer type bn_t detailed in relic_bn.h
"""
from relic import librelic
from common import *
from ctypes import Structure, byref, sizeof, c_int, c_ulonglong

class BigInt(Structure):
    """
    Multiple precision integer used by RELIC.
    """
    BN_SIZE = 34
    POSITIVE_FLAG = c_int(0)
    NEGATIVE_FLAG = c_int(1)

    # This maps to the type bn_st in relic_bn.h
    _fields_ = [
            ("alloc", c_int), 
            ("used", c_int), 
            ("sign", c_int), 
            # NOTE: We've promoted the type to c_ulonglong and halfed the 
            #       size of BN_SIZE compared to the representation in 
            #       relic_bn.h because BN_MAGNI = DOUBLE, meaning each "digit"
            #       is a double-word.
            ("digits", c_ulonglong * (BN_SIZE/2)), 
        ]


    def __init__(self, x=None):
        """
        Initialize an empty BigInt or create one from a Python long value @x
        """
        if x is None:
            return

        if not isinstance(x, (int, long)):
            raise TypeError("BigInt can only be initialized from a Python long or int value.")

        # The number of bits in each "digit"
        bits = sizeof(c_ulonglong)*8

        # Set the bookkeeping fields
        self.alloc = 0
        self.used = 0
        self.sign = BigInt.POSITIVE_FLAG if x >= 0 else BigInt.NEGATIVE_FLAG

        # Grab and set the "digits" from the value x. Digits are arranged
        # little-endian.
        x = abs(x)
        while x > 0:
            self.digits[self.used] = x % (2**bits)
            self.used += 1
            x >>= bits


    def __add__(self, other):
        """
        Add @other integer type to this BigInt.
        """
        other = coerceBigInt(other)
        if not other:
            return NotImplemented

        result = BigInt()
        librelic.bn_add(byref(result), byref(self), byref(other))
        return result


    def __eq__(self, other):
        """
        Compares this BigInt against an integer type.
        """
        return compare(self, other) == EQUAL


    def __le__(self, other):
        """
        Compares this BigInt against an integer type.
        """
        return compare(self, other) in [LESS_THAN, EQUAL]


    def __lt__(self, other):
        """
        Compares this BigInt against an integer type.
        """
        return compare(self, other) == LESS_THAN


    def __ge__(self, other):
        """
        Compares this BigInt against an integer type.
        """
        return compare(self, other) in [GREATER_THAN, EQUAL]


    def __gt__(self, other):
        """
        Compares this BigInt against an integer type.
        """
        return compare(self, other) == GREATER_THAN


    def __ne__(self, other):
        """
        Compares this BigInt against an @other BigInt or long value.
        """
        return not self.__eq__(other)


    def __str__(self):
        """
        Retrieves a hexadecimal representation of this BigInt. 
        """
        return "BigInt<{}>".format(hexString(self.digits[:self.used]))


    def __mod__(self, other):
        """
        Computes self modulo other where other is either a python integer
        type of a BigInt.
        """
        other = coerceBigInt(other)
        if not other:
            return NotImplemented

        result = BigInt()
        librelic.bn_mod_abi(byref(result), byref(self), byref(other))
        return result


    def __mul__(self, other):
        """
        Computes self*other assuming other is also a BigInt.
        """
        other = coerceBigInt(other)
        if not other: 
            return NotImplemented
        result = BigInt()
        librelic.bn_mul_basic(byref(result), byref(self), byref(other))
        return result


    def __sub__(self, other):
        """
        Subtract @other BigInt from this BigInt.
        """
        other = coerceBigInt(other)
        if not other:
            return NotImplemented

        result = BigInt()
        librelic.bn_sub(byref(result), byref(self), byref(other))
        return result


    def __long__(self):
        """
        Convert this BigInt to a Python long value.
        """
        # Add up the digits from smallest index to largest as specified in
        # relic_bn.h
        # r:result, d:digit
        r = long(0)
        for i,d in enumerate(self.digits):
            r += d << (sizeof(c_ulonglong)*8*i)
        return r


def compare(a, b):
    """
    Compares BigInt @a against integer type @b. Returns LESS_THAN, 
    EQUAL, or GREATER_THAN.
    """
    assertType(a, BigInt)
    b = coerceBigInt(b)
    if not b:
        return NotImplemented

    return librelic.bn_cmp(byref(a), byref(b))


def assertScalarType(x):
    """
    Ensures that @x is of type, BigInt, long, or int. Raises NotImplementedError
    otherwise.
    """
    assertType(x, (BigInt, int, long))


def coerceBigInt(x):
    """
    Retrieves a BigInt from @x or returns None if @x is not a type that can be
    converted.
    """
    # BigInt's are easy.
    if isinstance(x, BigInt):
        return x

    # Convert ints and longs using the constructor
    elif isinstance(x, (long, int)):
        return BigInt(x)

    else:
        return None


def hashZ(*args):
    """
    Hash @args into a BigInt using a cryptographic hash function.
    """
    TAG = "TAG_RELIC_HASH_Z"
    MESSAGE = "MESSAGE_HASH_Z"

    # Combine the arguments into a canonical string representation.
    text = TAG.join([str(val) for val in args ] )

    # Hash the string using HMAC
    # b: byte string 
    b = hmac(text, MESSAGE)
    return BigInt(longFromString(b))


def inverse(x, p, errorOnFail=False):
    """
    Find the inverse of BigInt @x in a field of (prime) order @p.
    """
    # Check types
    assertType(x, BigInt)

    # There are a number of ways in RELIC to compute this inverse, but
    # for simplicity, we'll use the extended GCD algorithm because it 
    # involves less type conversions. On the initial development platform
    # Lehmer's algorithm was the most performant: we call it directly.
    gcd = BigInt()
    inv = BigInt()

    # bn_gcd_ext(c, d, e, a, b) computes: c = a*d + b*e
    # We take x=a. b=p, and expect: c = 1 = gcd(x,p), d = 1/x, and e is unused.
    librelic.bn_gcd_ext_lehme(byref(gcd), byref(inv), None, byref(x), byref(p))

    # Check that GCD == 1 
    if gcd != 1:
        if errorOnFail:
            raise Exception("Cannot find an inverse. gcd(x,p) == {}, but we need gcd(x,p) == 1 to find an inverse.".
                format(long(gcd)))
        else:
            return None

    return inv


def randomZ(maximum=None, bits=256):
    """
    Retrieve a random BigInt.
    @maximum: If specified, the value will be no larger than this modulus.
    @bits: If no maximum is specified, the value will have @bits.
    """
    result = BigInt()

    # Select a random number smaller than the maximum.
    if maximum:
        maximum = coerceBigInt(maximum)
        librelic.bn_rand_mod(byref(result), byref(maximum))

    # Otherwise, select a random BigInt of the appropriate size in bits.
    else:
        librelic.bn_rand_abi(byref(result), BigInt.POSITIVE_FLAG, c_int(bits))
    
    return result
